@@ -1,12 +1,15 @@
-use crate::relay::{ProcessorRelay, WorkerRelay};
 use crate::tokio::{
     self,
     runtime::Runtime,
-    sync::mpsc::{channel, Sender},
+    sync::mpsc::{channel, Receiver, Sender},
     time::timeout,
 };
 use crate::{
-    error::Error, node::NullWorker, parser, relay::RelayMessage, Cancel, Mailbox, NodeMessage,
+    error::Error,
+    parser,
+    relay::{CtrlSignal, ProcessorRelay, RelayMessage, WorkerRelay},
+    router::SenderPair,
+    Cancel, NodeMessage, ShutdownType,
 };
 use core::time::Duration;
 use ockam_core::compat::{sync::Arc, vec::Vec};
@@ -36,7 +39,7 @@ pub struct Context {
     address: AddressSet,
     sender: Sender<NodeMessage>,
     rt: Arc<Runtime>,
-    mailbox: Mailbox,
+    mailbox: Receiver<RelayMessage>,
 }
 
 impl Context {
@@ -44,24 +47,37 @@ impl Context {
     pub fn runtime(&self) -> Arc<Runtime> {
         self.rt.clone()
     }
-    pub(crate) fn mailbox_mut(&mut self) -> &mut Mailbox {
-        &mut self.mailbox
+    /// Wait for the next message from the mailbox
+    pub(crate) async fn mailbox_next(&mut self) -> Option<RelayMessage> {
+        self.mailbox.recv().await
     }
 }
 
 impl Context {
+    /// Create a new context
+    ///
+    /// This function returns a new instance of Context, the relay
+    /// sender pair, and relay control signal receiver.
     pub(crate) fn new(
         rt: Arc<Runtime>,
         sender: Sender<NodeMessage>,
         address: AddressSet,
-        mailbox: Mailbox,
-    ) -> Self {
-        Self {
-            rt,
-            sender,
-            address,
-            mailbox,
-        }
+    ) -> (Self, SenderPair, Receiver<CtrlSignal>) {
+        let (mailbox_tx, mailbox) = channel(32);
+        let (ctrl_tx, ctrl_rx) = channel(1);
+        (
+            Self {
+                rt,
+                sender,
+                address,
+                mailbox,
+            },
+            SenderPair {
+                msgs: mailbox_tx,
+                ctrl: ctrl_tx,
+            },
+            ctrl_rx,
+        )
     }
 
     /// Return the primary worker address
@@ -86,11 +102,15 @@ impl Context {
     }
 
     async fn new_context_impl(&self, addr: Address) -> Result<Context> {
-        let (ctx, tx) = NullWorker::create(Arc::clone(&self.rt), &addr, self.sender.clone());
+        // Create a new context and get access to the mailbox senders
+        let (ctx, sender, _) = Self::new(
+            Arc::clone(&self.rt),
+            self.sender.clone(),
+            addr.clone().into(),
+        );
 
-        // Create a small relay and register it with the internal router
-        let sender = WorkerRelay::<NullWorker, _>::build_root(&self.rt, tx);
-        let (msg, mut rx) = NodeMessage::start_worker(addr.into(), sender);
+        // Create a "bare relay" and register it with the router
+        let (msg, mut rx) = NodeMessage::start_worker(addr.into(), sender, true);
         self.sender
             .send(msg)
             .await
@@ -118,29 +138,15 @@ impl Context {
         NM: Message + Send + 'static,
         NW: Worker<Context = Context, Message = NM>,
     {
-        // Check if the address set is available
-        // TODO: There is not much sense of checking for address collisions here, since in
-        // async environment there may be new Workers started between the check and actual adding
-        // of this Worker to the Router map, so check only should happen during Router::start_worker
-        let (check_addrs, mut check_rx) = NodeMessage::check_address(address.clone());
-        self.sender
-            .send(check_addrs)
-            .await
-            .map_err(|_| Error::InternalIOFailure)?;
-        check_rx.recv().await.ok_or(Error::InternalIOFailure)??;
-
-        // Build the mailbox first
-        let (mb_tx, mb_rx) = channel(32);
-        let mb = Mailbox::new(mb_rx);
-
         // Pass it to the context
-        let ctx = Context::new(self.rt.clone(), self.sender.clone(), address.clone(), mb);
+        let (ctx, sender, ctrl_rx) =
+            Context::new(self.rt.clone(), self.sender.clone(), address.clone());
 
         // Then initialise the worker message relay
-        let sender = WorkerRelay::<NW, NM>::build(self.rt.as_ref(), worker, ctx, mb_tx);
+        WorkerRelay::<NW, NM>::init(self.rt.as_ref(), worker, ctx, ctrl_rx);
 
         // Send start request to router
-        let (msg, mut rx) = NodeMessage::start_worker(address, sender);
+        let (msg, mut rx) = NodeMessage::start_worker(address, sender, false);
         self.sender
             .send(msg)
             .await
@@ -166,24 +172,16 @@ impl Context {
     where
         P: Processor<Context = Context>,
     {
-        // Build the mailbox first
-        let (mb_tx, mb_rx) = channel(32);
-        let mb = Mailbox::new(mb_rx);
+        let addr = address.clone();
 
-        // Pass it to the context
-        let ctx = Context::new(
-            self.rt.clone(),
-            self.sender.clone(),
-            address.clone().into(),
-            mb,
-        );
+        let (ctx, senders, ctrl_rx) =
+            Context::new(self.rt.clone(), self.sender.clone(), addr.into());
 
-        // Then initialise the worker message relay
-        let (sender, shutdown_handle) =
-            ProcessorRelay::<P>::build(self.rt.as_ref(), processor, ctx, mb_tx);
+        // Initialise the processor relay with the ctrl receiver
+        ProcessorRelay::<P>::init(self.rt.as_ref(), processor, ctx, ctrl_rx);
 
         // Send start request to router
-        let (msg, mut rx) = NodeMessage::start_processor(address, sender, shutdown_handle);
+        let (msg, mut rx) = NodeMessage::start_processor(address, senders);
         self.sender
             .send(msg)
             .await
@@ -225,16 +223,48 @@ impl Context {
             .map(|_| ())?)
     }
 
-    /// Signal to the local application runner to shut down
-    pub async fn stop(&mut self) -> Result<()> {
+    /// Signal to the local runtime to shut down immediately
+    ///
+    /// **WARNING**: calling this function may result in data loss.
+    /// It is recommended to use the much safer
+    /// [`Context::stop`](Context::stop) function instead!
+    pub async fn stop_now(&mut self) -> Result<()> {
         let tx = self.sender.clone();
         info!("Shutting down all workers");
-        match tx.send(NodeMessage::StopNode).await {
+        let (msg, _) = NodeMessage::stop_node(ShutdownType::Immediate);
+
+        match tx.send(msg).await {
             Ok(()) => Ok(()),
             Err(_e) => Err(Error::FailedStopNode.into()),
         }
     }
 
+    /// Signal to the local runtime to shut down
+    ///
+    /// This call will hang until a safe shutdown has been completed.
+    /// The default timeout for a safe shutdown is 1 second.  You can
+    /// change this behaviour by calling
+    /// [`Context::stop_timeout`](Context::stop_timeout) directly.
+    pub async fn stop(&mut self) -> Result<()> {
+        self.stop_timeout(1).await
+    }
+
+    /// Signal to the local runtime to shut down
+    ///
+    /// This call will hang until a safe shutdown has been completed
+    /// or the desired timeout has been reached.
+    pub async fn stop_timeout(&mut self, seconds: u8) -> Result<()> {
+        let (req, mut rx) = NodeMessage::stop_node(ShutdownType::Graceful(seconds));
+        self.sender.send(req).await.map_err(Error::from)?;
+
+        // Wait until we get the all-clear
+        Ok(rx
+            .recv()
+            .await
+            .ok_or(Error::InternalIOFailure)?
+            .map(|_| ())?)
+    }
+
     /// Send a message via a fully qualified route
     ///
     /// Routes can be constructed from a set of [`Address`]es, or via
@@ -423,6 +453,27 @@ impl Context {
         Ok(Cancel::new(m, data, addr, self))
     }
 
+    /// Assign the current worker to a cluster
+    ///
+    /// A cluster is a set of workers that should be stopped together
+    /// when the node is stopped or parts of the system are reloaded.
+    /// **This is not to be confused with supervisors!**
+    ///
+    /// By adding your worker to a cluster you signal to the runtime
+    /// that your worker may be depended on by other workers that
+    /// should be stopped first.
+    ///
+    /// **Your cluster name MUST NOT start with `_internals.` or
+    /// `ockam.`!**
+    ///
+    /// Clusters are de-allocated in reverse order of their
+    /// initialisation when the node is stopped.
+    pub async fn set_cluster<S: Into<String>>(&self, label: S) -> Result<()> {
+        let (msg, mut rx) = NodeMessage::set_cluster(self.address(), label.into());
+        self.sender.send(msg).await.map_err(Error::from)?;
+        Ok(rx.recv().await.ok_or(Error::InternalIOFailure)??.is_ok()?)
+    }
+
     /// Return a list of all available worker addresses on a node
     pub async fn list_workers(&self) -> Result<Vec<Address>> {
         let (msg, mut reply_rx) = NodeMessage::list_workers();
@@ -441,6 +492,15 @@ impl Context {
         self.register_impl(type_, addr.into()).await
     }
 
+    /// Send a shutdown acknowlegement to the router
+    pub(crate) async fn send_stop_ack(&self) -> Result<()> {
+        self.sender
+            .send(NodeMessage::StopAck(self.address()))
+            .await
+            .map_err(Error::from)?;
+        Ok(())
+    }
+
     async fn register_impl(&self, type_: u8, addr: Address) -> Result<()> {
         let (tx, mut rx) = channel(1);
         self.sender
@@ -466,7 +526,7 @@ impl Context {
     /// has woken it.
     async fn next_from_mailbox<M: Message>(&mut self) -> Result<(M, LocalMessage, Address)> {
         loop {
-            let msg = self.mailbox.next().await.ok_or(Error::FailedLoadData)?;
+            let msg = self.mailbox_next().await.ok_or(Error::FailedLoadData)?;
             let (addr, data) = msg.local_msg();
 
             // FIXME: make message parsing idempotent to avoid cloning
