@@ -1,6 +1,5 @@
-use crate::relay::ShutdownHandle;
 use crate::tokio::sync::mpsc::{channel, Receiver, Sender};
-use crate::{error::Error, relay::RelayMessage};
+use crate::{error::Error, relay::RelayMessage, router::SenderPair};
 use ockam_core::compat::vec::Vec;
 use ockam_core::{Address, AddressSet};
 
@@ -8,51 +7,71 @@ use ockam_core::{Address, AddressSet};
 #[derive(Debug)]
 pub enum NodeMessage {
     /// Start a new worker and store the send handle
-    StartWorker(AddressSet, Sender<RelayMessage>, Sender<NodeReplyResult>),
+    StartWorker {
+        /// The set of addresses in use by this worker
+        addrs: AddressSet,
+        /// Pair of senders to the worker relay (msgs and ctrl)
+        senders: SenderPair,
+        /// A bare worker runs no relay state
+        bare: bool,
+        /// Reply channel for command confirmation
+        reply: Sender<NodeReplyResult>,
+    },
     /// Return a list of all worker addresses
     ListWorkers(Sender<NodeReplyResult>),
+    /// Add an existing address to a cluster
+    SetCluster(Address, String, Sender<NodeReplyResult>),
     /// Stop an existing worker
     StopWorker(Address, Sender<NodeReplyResult>),
-    /// Start a new processor and store the send and shutdown handles
-    StartProcessor(
-        Address,
-        Sender<RelayMessage>,
-        Sender<NodeReplyResult>,
-        ShutdownHandle,
-    ),
+    /// Start a new processor
+    StartProcessor(Address, SenderPair, Sender<NodeReplyResult>),
     /// Stop an existing processor
     StopProcessor(Address, Sender<NodeReplyResult>),
     /// Stop the node (and all workers)
-    StopNode,
+    StopNode(ShutdownType, Sender<NodeReplyResult>),
+    /// Immediately stop the node runtime
+    AbortNode,
+    /// Let the router know a particular address has stopped
+    StopAck(Address),
     /// Request the sender for a worker address
     SenderReq(Address, Sender<NodeReplyResult>),
     /// Register a new router for a route id type
     Router(u8, Address, Sender<NodeReplyResult>),
-    /// Check if a given address is already registered
-    CheckAddress(AddressSet, Sender<NodeReplyResult>),
 }
 
 impl NodeMessage {
     /// Create a start worker message
+    ///
+    /// * `senders`: message and command senders for the relay
+    ///
+    /// * `bare`: indicate whether this worker address has a full
+    ///   relay behind it that can respond to shutdown commands.
+    ///   Setting this to `true` will disable stop ACK support in the
+    ///   router
     pub fn start_worker(
-        address: AddressSet,
-        sender: Sender<RelayMessage>,
+        addrs: AddressSet,
+        senders: SenderPair,
+        bare: bool,
     ) -> (Self, Receiver<NodeReplyResult>) {
-        let (tx, rx) = channel(1);
-        (Self::StartWorker(address, sender, tx), rx)
+        let (reply, rx) = channel(1);
+        (
+            Self::StartWorker {
+                addrs,
+                senders,
+                bare,
+                reply,
+            },
+            rx,
+        )
     }
 
     /// Create a start worker message
     pub fn start_processor(
         address: Address,
-        sender: Sender<RelayMessage>,
-        shutdown_handle: ShutdownHandle,
+        senders: SenderPair,
     ) -> (Self, Receiver<NodeReplyResult>) {
         let (tx, rx) = channel(1);
-        (
-            Self::StartProcessor(address, sender, tx, shutdown_handle),
-            rx,
-        )
+        (Self::StartProcessor(address, senders, tx), rx)
     }
 
     /// Create a stop worker message and reply receiver
@@ -67,6 +86,12 @@ impl NodeMessage {
         (Self::ListWorkers(tx), rx)
     }
 
+    /// Create a set cluster message and reply receiver
+    pub fn set_cluster(addr: Address, label: String) -> (Self, Receiver<NodeReplyResult>) {
+        let (tx, rx) = channel(1);
+        (Self::SetCluster(addr, label, tx), rx)
+    }
+
     /// Create a stop worker message and reply receiver
     pub fn stop_worker(address: Address) -> (Self, Receiver<NodeReplyResult>) {
         let (tx, rx) = channel(1);
@@ -74,8 +99,9 @@ impl NodeMessage {
     }
 
     /// Create a stop node message
-    pub fn stop_node() -> Self {
-        Self::StopNode
+    pub fn stop_node(tt: ShutdownType) -> (Self, Receiver<NodeReplyResult>) {
+        let (tx, rx) = channel(1);
+        (Self::StopNode(tt, tx), rx)
     }
 
     /// Create a sender request message and reply receiver
@@ -83,12 +109,6 @@ impl NodeMessage {
         let (tx, rx) = channel(1);
         (Self::SenderReq(route, tx), rx)
     }
-
-    /// Create a message to check the availability of an address set
-    pub fn check_address(addrs: AddressSet) -> (Self, Receiver<NodeReplyResult>) {
-        let (tx, rx) = channel(1);
-        (Self::CheckAddress(addrs, tx), rx)
-    }
 }
 
 /// The reply/result of a Node
@@ -124,6 +144,60 @@ pub enum NodeError {
     WorkerExists(Address),
     /// Router already exists
     RouterExists,
+    /// Command rejected
+    Rejected(Reason),
+}
+
+/// The reason why a command was rejected
+#[derive(Debug, Copy, Clone)]
+pub enum Reason {
+    /// Rejected because the node is currently shutting down
+    NodeShutdown,
+    /// Rejected because the worker is currently shutting down
+    WorkerShutdown,
+}
+
+/// Specify the type of node shutdown
+///
+/// For most users `ShutdownType::Graceful()` is recommended.  The
+/// `Default` implementation uses a 1 second timeout.
+#[derive(Debug, Copy, Clone)]
+#[non_exhaustive]
+pub enum ShutdownType {
+    /// Execute a graceful shutdown given a maximum timeout
+    ///
+    /// The following steps will be taken by the internal router
+    /// during graceful shutdown procedure:
+    ///
+    /// * Signal clusterless workers to stop
+    /// * Wait for shutdown ACK hooks from worker set
+    /// * Signal worker clusters in reverse-creation order to stop
+    /// * Wait for shutdown ACK hooks from each cluster before moving onto the
+    ///   next
+    /// * All shutdown-signalled workers may process their entire mailbox,
+    ///   while not allowing new messages to be queued
+    ///
+    /// Graceful shutdown procedure will be pre-maturely terminated
+    /// when reaching the timeout (failover into `Immediate`
+    /// strategy).  **A given timeout of `0` will wait forever!**
+    Graceful(u8),
+    /// Immediately shutdown workers and run shutdown hooks
+    ///
+    /// This strategy can lead to data loss:
+    ///
+    /// * Unhandled mailbox messages will be dropped
+    /// * Shutdown hooks may not be able to send messages
+    ///
+    /// This strategy is not recommended for general use, but will be
+    /// selected as a failover, if the `Graceful` strategy reaches its
+    /// timeout limit.
+    Immediate,
+}
+
+impl Default for ShutdownType {
+    fn default() -> Self {
+        Self::Graceful(1)
+    }
 }
 
 impl NodeReply {
@@ -152,6 +226,11 @@ impl NodeReply {
         Err(NodeError::RouterExists)
     }
 
+    /// Return [NodeReply::Rejected(reason)]
+    pub fn rejected(reason: Reason) -> NodeReplyResult {
+        Err(NodeError::Rejected(reason))
+    }
+
     /// Return [NodeReply::Workers] for the given addresses
     pub fn workers(v: Vec<Address>) -> NodeReplyResult {
         Ok(Self::Workers(v))
